import { BEFORE_REQUEST, AFTER_REQUEST, INTERRUPTED } from './constants.js';
import { validateMiddleware } from './validateMiddleware.js';

export { INTERRUPTED };

export class Middleware {
  constructor() {
    this._middlewares = [];
    this._requests = {};

    this.executePreRequest = this.executePreRequest.bind(this);
    this.executePostRequest = this.executePostRequest.bind(this);
    this._proceed = this._proceed.bind(this);
    this._quit = this._quit.bind(this);
  }

  getAll() {
    return this._middlewares;
  }

  use(middleware) {
    this._middlewares.push(validateMiddleware(middleware));
    return this;
  }

  executePreRequest(requestId, request) {
    return () => this._executeStage(BEFORE_REQUEST, requestId, request);
  }

  executePostRequest(requestId, request) {
    return response => this._executeStage(
      AFTER_REQUEST, requestId, request, response);
  }

  _executeStage(stage, requestId, request, response) {
    this._requests[requestId] = this._requests[requestId] || {};
    this._requests[requestId].stage = stage;
    this._requests[requestId].response = response;

    this._executeCurrentMiddleware(requestId, request);

    return new Promise((resolve) => {
      this._requests[requestId].resolver = (signal) => {
        const userResponse = this._requests[requestId].response;
        resolve(signal || userResponse || response);
      };
    });
  }

  _executeCurrentMiddleware(requestId, request) {
    const { stage, response, index = 0 } = this._requests[requestId];
    const currentMiddleware = this._middlewares[index];

    if (!currentMiddleware || !currentMiddleware[stage]) {
      this._proceed(requestId, request);
      return undefined;
    }

    const currentCallback = currentMiddleware[stage];

    const signal = stage === BEFORE_REQUEST
      ? INTERRUPTED
      : undefined;

    currentCallback({
      proceed: (userResponse) => {
        this._requests[requestId] = this._requests[requestId] || {};
        this._requests[requestId].response = userResponse || response;
        this._proceed(requestId, request);
      },
      quit: () => {
        this._quit(requestId, signal);
      },
      request,
      response,
    });

    return undefined;
  }

  _proceed(requestId, request) {
    const { index = 0 } = this._requests[requestId];

    if (index + 1 < this._middlewares.length) {
      this._requests[requestId].index = index + 1;
      this._executeCurrentMiddleware(requestId, request);
      return undefined;
    }

    this._quit(requestId);
    return undefined;
  }

  _quit(requestId, signal) {
    Promise.resolve().then(() => {
      this._requests[requestId].resolver(signal);
      this._requests[requestId] = undefined;
    });
  }
}

export const middleware = new Middleware();
