import sinon from 'sinon';
import chai from 'chai';
import sinonChai from 'sinon-chai';

import { Middleware, INTERRUPTED } from './middleware.js';
import { uniqId } from '../../helpers/uniqId.js';

chai.use(sinonChai);
const { expect } = chai;

let middleware;
let requestId;
let requestIds;

const invalidMiddleware = null;

const validMiddleware = {
  beforeRequest() {},
  afterRequest() {},
};

const delay = time => new Promise(resolve => setTimeout(resolve, time));

describe('Middleware Service', () => {
  beforeEach(() => {
    middleware = new Middleware();
    requestId = uniqId();
    requestIds = [uniqId(), uniqId(), uniqId()];
  });

  afterEach(() => {
    middleware = undefined;
    requestId = undefined;
    requestIds = undefined;
  });

  describe('#getAll()', () => {
    it('Should expose all middlewares', () => {
      middleware._middlewares = [validMiddleware];
      expect(middleware.getAll()).to.deep.equal([validMiddleware]);
    });
  });

  describe('#use()', () => {
    it('Should queue a valid middleware', () => {
      middleware.use(validMiddleware);
      expect(middleware._middlewares.length).to.equal(1);

      middleware.use(validMiddleware);
      expect(middleware._middlewares.length).to.equal(2);

      middleware.use(validMiddleware);
      expect(middleware._middlewares.length).to.equal(3);
    });

    it('Should throw an error when receiveing an invalid middleware', () => {
      expect(() => middleware.use(invalidMiddleware)).to.throw();
    });
  });

  describe('#executePreRequest()', () => {
    beforeEach(() => {
      middleware._executeCurrentMiddleware = sinon.spy();
    });

    it('Should return a function', () => {
      expect(middleware.executePreRequest()).to.be.a('function');
    });

    it('Should run _executeCurrentMiddleware', () => {
      middleware.executePreRequest(requestId)();
      expect(middleware._executeCurrentMiddleware).to.have.been.calledOnce;
    });

    it('Should return promise', () => {
      expect(middleware.executePreRequest(requestId)()).to.be.a('promise');
    });

    it('Should return the original argument when resolved', () => {
      const ARG = Symbol('ARG');

      middleware.executePreRequest(requestId)(ARG)
        .then((resolved) => {
          expect(resolved).to.equal(ARG);
        });

      middleware._requests[requestId].resolver();
    });

    it('Should return a signal when resolved with one', () => {
      const ARG = Symbol('ARG');

      middleware.executePreRequest(requestId)(ARG)
        .then((resolved) => {
          expect(resolved).to.equal(INTERRUPTED);
        });

      middleware._requests[requestId].resolver(INTERRUPTED);
    });
  });

  describe('#executePostRequest()', () => {
    beforeEach(() => {
      middleware._executeCurrentMiddleware = sinon.spy();
    });

    it('Should return a function', () => {
      expect(middleware.executePostRequest()).to.be.a('function');
    });

    it('Should run _executeCurrentMiddleware', () => {
      middleware.executePostRequest(requestId)();
      expect(middleware._executeCurrentMiddleware).to.have.been.calledOnce;
    });

    it('Should return a pending promise', () => {
      expect(middleware.executePostRequest(requestId)()).to.be.a('promise');
    });

    it('Should return the original argument when resolved', () => {
      const ARG = Symbol('ARG');

      middleware.executePostRequest(requestId)(ARG)
        .then((resolved) => {
          expect(resolved).to.equal(ARG);
        });

      middleware._requests[requestId].resolver();
    });

    it('Should return a signal when resolved with one', () => {
      const ARG = Symbol('ARG');

      middleware.executePostRequest(requestId)(ARG)
        .then((resolved) => {
          expect(resolved).to.equal(INTERRUPTED);
        });

      middleware._requests[requestId].resolver(INTERRUPTED);
    });
  });

  it('Should execute all beforeRequest actions ' +
    'that call proceed, in order', async () => {
    const beforeOneSpy = sinon.spy();
    const beforeTwoSpy = sinon.spy();
    const afterOneSpy = sinon.spy();

    const midOne = {
      async beforeRequest({ proceed }) {
        await delay(50);
        beforeOneSpy();
        proceed();
      },
      afterRequest({ proceed }) {
        afterOneSpy();
        proceed();
      },
    };

    const midTwo = {
      beforeRequest({ proceed }) {
        beforeTwoSpy();
        proceed();
      },
    };

    middleware.use(midOne);
    middleware.use(midTwo);

    await middleware.executePreRequest(requestId)();

    expect(beforeOneSpy).to.have.been.calledOnce;
    expect(beforeTwoSpy).to.have.been.calledOnce;
    expect(beforeTwoSpy).to.have.been.calledAfter(beforeOneSpy);
    expect(afterOneSpy).to.not.have.been.called;
  });

  it('Should execute all beforeRequest actions ' +
    'until quit is called', async () => {
    const beforeOneSpy = sinon.spy();
    const beforeTwoSpy = sinon.spy();
    const afterOneSpy = sinon.spy();

    const midOne = {
      async beforeRequest({ quit }) {
        await delay(50);
        beforeOneSpy();
        quit();
      },
      afterRequest({ proceed }) {
        afterOneSpy();
        proceed();
      },
    };

    const midTwo = {
      beforeRequest({ proceed }) {
        beforeTwoSpy();
        proceed();
      },
    };

    middleware.use(midOne);
    middleware.use(midTwo);

    await middleware.executePreRequest(requestId)();

    expect(beforeOneSpy).to.have.been.calledOnce;
    expect(beforeTwoSpy).to.not.have.been.called;
    expect(afterOneSpy).to.not.have.been.called;
  });

  it('Should execute all afterRequest actions ' +
    'that call proceed, in order', async () => {
    const beforeOneSpy = sinon.spy();
    const beforeTwoSpy = sinon.spy();
    const afterOneSpy = sinon.spy();
    const afterTwoSpy = sinon.spy();

    const midOne = {
      async beforeRequest({ proceed }) {
        await delay(50);
        beforeOneSpy();
        proceed();
      },
      afterRequest({ proceed }) {
        afterOneSpy();
        proceed();
      },
    };

    const midTwo = {
      beforeRequest({ proceed }) {
        beforeTwoSpy();
        proceed();
      },
    };

    const midThree = {
      afterRequest({ proceed }) {
        afterTwoSpy();
        proceed();
      },
    };

    middleware.use(midOne);
    middleware.use(midTwo);
    middleware.use(midThree);

    await middleware.executePostRequest(requestId)();

    expect(beforeOneSpy).to.not.have.been.called;
    expect(beforeTwoSpy).to.not.have.been.called;
    expect(afterOneSpy).to.have.been.calledOnce;
    expect(afterTwoSpy).to.have.been.calledOnce;
    expect(afterTwoSpy).to.have.been.calledAfter(afterOneSpy);
  });

  it('Should execute all afterRequest actions ' +
    'until quit is called', async () => {
    const beforeOneSpy = sinon.spy();
    const beforeTwoSpy = sinon.spy();
    const afterOneSpy = sinon.spy();
    const afterTwoSpy = sinon.spy();

    const midOne = {
      async beforeRequest({ quit }) {
        await delay(50);
        beforeOneSpy();
        quit();
      },
      afterRequest({ quit }) {
        afterOneSpy();
        quit();
      },
    };

    const midTwo = {
      beforeRequest({ proceed }) {
        beforeTwoSpy();
        proceed();
      },
    };

    const midThree = {
      afterRequest({ proceed }) {
        afterTwoSpy();
        proceed();
      },
    };

    middleware.use(midOne);
    middleware.use(midTwo);
    middleware.use(midThree);

    await middleware.executePostRequest(requestId)();

    expect(beforeOneSpy).to.not.have.been.called;
    expect(beforeTwoSpy).to.not.have.been.called;
    expect(afterOneSpy).to.have.been.calledOnce;
    expect(afterTwoSpy).to.not.have.been.called;
  });

  it('Should start from the beginning of the queue every time', async () => {
    const beforeOneSpy = sinon.spy();
    const beforeTwoSpy = sinon.spy();

    const midOne = {
      async beforeRequest({ proceed }) {
        await delay(50);
        beforeOneSpy();
        proceed();
      },
    };

    const midTwo = {
      beforeRequest({ proceed }) {
        beforeTwoSpy();
        proceed();
      },
    };

    middleware.use(midOne);
    middleware.use(midTwo);

    await middleware.executePreRequest(requestIds[0])();
    await middleware.executePreRequest(requestIds[1])();

    expect(beforeOneSpy).to.have.been.calledTwice;
    expect(beforeTwoSpy).to.have.been.calledTwice;

    expect(beforeOneSpy).to.have.been.calledAfter(beforeTwoSpy);
    expect(beforeTwoSpy).to.have.been.calledAfter(beforeOneSpy);
  });

  it('Should not break when called simultaneously', async () => {
    const beforeOneSpy = sinon.spy();

    const midOne = {
      async beforeRequest({ proceed }) {
        await delay(50);
        beforeOneSpy();
        proceed();
      },
    };

    middleware.use(midOne);

    await Promise.all([
      middleware.executePreRequest(requestIds[0])(),
      middleware.executePreRequest(requestIds[1])(),
      middleware.executePreRequest(requestIds[2])(),
    ]);

    expect(beforeOneSpy).to.have.been.calledThrice;
  });

  it('Should allow middlewares to be chainable', () => {
    expect(middleware.use({})).to.equal(middleware);
  });

  it('Should allow user to override request', async () => {
    const midOne = {
      async afterRequest({ proceed }) {
        proceed({ overriden: true });
      },
    };

    middleware.use(midOne);

    const result = await middleware.executePostRequest(requestIds[0])();

    expect(result).to.deep.equal({ overriden: true });
  });
});
